import pytest

from . import utils
from conjurer import ml


@pytest.fixture
def setup():
    df_training = utils.get_input_df(100)
    df_validation = utils.get_input_df(100)
    df_test = utils.get_input_df(10)
    feature_columns = ["column{}".format(i) for i in range(6)]
    return df_training, df_validation, df_test, feature_columns


def test_grid_cv(setup):
    df_training, df_validation, df_test, feature_columns = setup
    is_cl = True
    target_column = utils.get_target_column(is_cl)
    cv_obj = ml.get_default_cv("linear_model", "cl", search_type="grid")
    cv_obj.parameter_grid={
        "ml__penalty": ["l1", "l2"],
        "ml__C": [1e-5, 1e-3, 1e-1]
    }
    model = cv_obj.fit_cv_pandas(df_training, target_column, feature_columns, n_fold=3)
    analyzer = ml.CVAnalyzer(model.estimator)
    _basic_flow(analyzer)


def test_random_cv(setup):
    df_training, df_validation, df_test, feature_columns = setup
    is_cl = False
    target_column = utils.get_target_column(is_cl)
    cv_obj = ml.get_default_cv("linear_model", "rg", "r2")
    model = cv_obj.fit_cv_pandas(df_training, target_column, feature_columns, n_fold=3)
    analyzer = ml.CVAnalyzer(model.estimator)
    _basic_flow(analyzer)


def test_grid_sv(setup):
    df_training, df_validation, df_test, feature_columns = setup
    is_cl = True
    target_column = utils.get_target_column(is_cl)
    cv_obj = ml.get_default_cv("linear_model", "cl", search_type="grid")
    cv_obj.parameter_grid={
        "ml__penalty": ["l1", "l2"],
        "ml__C": [1e-5, 1e-3, 1e-1]
    }
    model = cv_obj.fit_sv_pandas(df_training, target_column, feature_columns, ratio_training=0.8)
    analyzer = ml.CVAnalyzer(model.estimator)
    _basic_flow(analyzer)


def test_random_sv(setup):
    df_training, df_validation, df_test, feature_columns = setup
    is_cl = False
    target_column = utils.get_target_column(is_cl)
    cv_obj = ml.get_default_cv("linear_model", "rg", "r2")
    model = cv_obj.fit_sv_pandas(df_training, target_column, feature_columns, df_validation=df_validation)
    analyzer = ml.CVAnalyzer(model.estimator)
    _basic_flow(analyzer)


def _basic_flow(analyzer):
    analyzer.plot_flat()
    analyzer.plot_by_param_all()
    analyzer.get_summary()
    analyzer.get_best_model_info()
    analyzer.get_param_stat_df()



import logging

import numpy
import pandas


logger = logging.getLogger(__name__)


class Model(object):
    """
    Class to provide sklearn model with prediction interface with pandas.DataFrame
    """
    def __init__(self, estimator, feature_columns, target_column=None):
        """
        Args:
            estimator: sklearn estimator / pipeline
            feature_columns (list of str): column names used for features
            target_column (str): column name used for target column
        """
        self.estimator = estimator
        self.feature_columns = feature_columns
        self.target_column = target_column
        self.model = get_ml_model_from_pipeline(get_estimator_from_cvobj(estimator))
        self.is_cl = hasattr(estimator, "predict_proba")

    def predict(self, df):
        index = df.index
        x = self.get_x_from_df(df)
        result_df = self._get_result_for_cl(x, index) if self.is_cl else self._get_result_for_rg(x, index)
        other_columns = [c for c in df.columns if c not in self.feature_columns]
        for c in other_columns:
            result_df[c] = df[c]
        return result_df

    def get_x_from_df(self, df):
        return df[self.feature_columns].values

    def get_y_from_df(self, df):
        return df[self.target_column].values

    def _get_result_for_rg(self, x, index):
        return pandas.DataFrame({
            "score": self.estimator.predict(x)
        }, index=index)

    def _get_result_for_cl(self, x, index):
        predicted_class = self.estimator.predict(x)
        proba_array = self.estimator.predict_proba(x)
        num_class = proba_array.shape[1]
        if num_class == 2:
            return pandas.DataFrame({
                "score": proba_array[:, 1],
                "predicted_class": predicted_class
            }, index=index, columns=["score", "predicted_class"])
        else:
            return pandas.DataFrame({
                **{
                    "score_{}".format(i): proba_array[:, i]
                    for i in range(num_class)
                },
                **{"predicted_class": predicted_class}
            }, index=index, columns=["score_{}".format(i) for i in range(num_class)] + ["predicted_class"])

    def get_feature_importance(self):
        return get_linear_importance(self.model, self.feature_columns) if hasattr(self.model, "coef_") \
            else get_feature_importance_sklearn(self.model, self.feature_columns)


def get_linear_importance(model, feature_columns):
    coeff_df = pandas.DataFrame({
        "feature_name": feature_columns,
        "importance": numpy.abs(model.coef_[0, :]),
        "coeff": model.coef_[0, :]
    })
    coeff_df = coeff_df.sort_values(by="importance", ascending=False)
    coeff_df["rank"] = range(len(coeff_df))
    return coeff_df


def get_feature_importance_sklearn(model, feature_columns):
    fi_df = pandas.DataFrame({
        "feature_name": feature_columns,
        "importance": model.feature_importances_
    }).sort_values(by="importance", ascending=False)
    fi_df["rank"] = range(len(fi_df))
    return fi_df


def get_estimator_from_cvobj(estimator):
    return estimator.best_estimator_ if hasattr(estimator, "best_estimator_") else estimator


def get_ml_model_from_pipeline(estimator):
    return estimator.steps[-1][1] if hasattr(estimator, "steps") else estimator



import os
import gc
import subprocess
import logging


logger = logging.getLogger(__name__)


def execute_command(command_list, show_stdout=False, show_stderr=False, show_command=False, no_fork=False, stdin=None,
                    file_input=None):
    try:
        if no_fork:
            os.system(" ".join(command_list))
        else:
            stdout, stderr = subprocess_execute(command_list, stdin, file_input)
            if show_stdout:
                logger.info("stdout:\n{stdout}".format(stdout=stdout))
            if show_stderr:
                logger.info("stderr:\n{stderr}".format(stderr=stderr))
    except:
        raise
    finally:
        if show_command:
            logger.info("command:\n{command}".format(command=" ".join(command_list)))


def subprocess_execute(command_list, std_input=None, file_input=None):
    gc.collect()
    popen_obj = subprocess.Popen(command_list, stdin=subprocess.PIPE, 
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    if file_input: popen_obj.stdin.write('@{0}'.format(file_input))
    if std_input: popen_obj.stdin.write(std_input)
    stdout, stderr = popen_obj.communicate()
    return_code = popen_obj.wait()
    if return_code != 0:
        error_message = get_error_message(command_list, stdout, stderr)
        logger.error(error_message)
        raise Exception(error_message)
    return stdout, stderr


def get_error_message(command_list, stdout="", stderr=""):
    error_message = \
'''
command execution failed:
    command:
{command}
'''.format(command=" ".join(command_list))
    if stdout:
        error_message += \
'''
    stdout:
{stdout}
'''.format(stdout=stdout)
    if stderr:
        error_message += \
'''
    stderr:
{stderr}
'''.format(stderr=stderr)
    return error_message



import logging

import numpy
import pandas
from pandas.api import types


logger = logging.getLogger(__name__)


def create_frequency_table(series, num_bins=50, minv=None, maxv=None):
    if is_quantitative(series, num_bins):
        return count_frequency_quantitative(series, num_bins, minv, maxv)
    else:
        return count_frequency_categorical(series, num_bins)


def is_quantitative(series, num_bins):
    if types.is_integer_dtype(series.dtype):
        return len(series.value_counts()) >= num_bins
    else:
        return types.is_numeric_dtype(series.dtype) or types.is_datetime64_any_dtype(series.dtype)


def count_frequency_quantitative(series, num_bins, minv, maxv):
    name = series.name or "value"
    bins = create_bins_quantitative(series, num_bins, minv, maxv)
    count_array = get_continuous_value_counts(series.values, bins)
    ratio_array = count_array / count_array.sum()
    return pandas.DataFrame({
        "{}_lb".format(name): [b.lb for b in bins],
        "{}_ub".format(name): [b.ub for b in bins],
        "frequency": count_array,
        "ratio": ratio_array
    })


def count_frequency_categorical(series, num_bins):
    name = series.name or "value"
    vcounts = series.value_counts()
    if len(vcounts) > num_bins:
        index = vcounts[:num_bins].index
        vcounts = get_categorical_value_counts(series, index)
    total = vcounts.values.sum()
    return pandas.DataFrame({
        name: vcounts.index,
        "frequency": vcounts.values,
        "ratio": [v/total for v in vcounts.values]
    })


def create_bins_quantitative(series, num_bins, minv=None, maxv=None):
    array = series.values
    if numpy.array(array[~numpy.isnan(array)]).size == 0:
        raise BinCreationError("all values are null")
    minv = minv if minv is not None else series.min()  # cannot use "or" for the case minv==0
    maxv = maxv if maxv is not None else series.max()  # cannot use "or" for the case maxv==0
    if maxv == minv:
        raise BinCreationError("min == max")
    lb_list, ub_list, bin_size = get_bin_config(minv, maxv, num_bins)
    return [QuantitativeBin(lb, ub, False) for lb, ub in zip(lb_list[:-1], ub_list[:-1])] + \
           [QuantitativeBin(lb_list[-1], ub_list[-1], True)]


def create_bins_categorical(series, num_bins):
    vcounts = series.value_counts()
    bins = [CategoricalBin([v]) for v in vcounts[:num_bins].index]
    if len(vcounts) > num_bins:
        bins.append(CategoricalBin([v for v in vcounts[num_bins:].index]))
    return bins


def get_bin_config(minv, maxv, num_bins):
    bin_size = (maxv - minv) / num_bins
    lower_list = [minv + i * bin_size for i in range(num_bins)]
    upper_list = [minv + (i + 1) * bin_size for i in range(num_bins-1)] + [maxv]
    return lower_list, upper_list, bin_size


def get_continuous_value_counts(series, bins):
    return numpy.array([len(b.filter(series)) for b in bins])


def get_categorical_value_counts(series, index=None):
    vcounts = series.value_counts()
    if index is not None:
        other_counts = vcounts[~vcounts.index.isin(index)].values.sum()
        vcounts = vcounts[index]
        vcounts["OTHER"] = other_counts
    return vcounts


def create_frequency_table_2d(df, column_x, column_y, num_bins_x=50, num_bins_y=50,
                              xmin=None, xmax=None, ymin=None, ymax=None):
    is_quantitative_x = is_quantitative(df[column_x], num_bins_x)
    is_quantitative_y = is_quantitative(df[column_y], num_bins_y)
    bins_x = create_bins(df[column_x], num_bins_x, xmin, xmax, is_quantitative_x)
    bins_y = create_bins(df[column_y], num_bins_y, ymin, ymax, is_quantitative_y)
    frequency_list = []
    for bin_x in bins_x:
        for bin_y in bins_y:
            frequency_list.append(
                _count_2d(df, column_x, column_y, bin_x, bin_y)
            )
    count_df = pandas.concat(frequency_list, ignore_index=True)
    count_df["ratio"] = count_df["frequency"] / count_df["frequency"].sum()
    return count_df


def create_bins(series, num_bins, minv, maxv, is_quantitative):
    if is_quantitative:
        return create_bins_quantitative(series, num_bins, minv, maxv)
    else:
        return create_bins_categorical(series, num_bins)


def _count_2d(df, column_x, column_y, bin_x, bin_y):
    return pandas.DataFrame({
        **bin_x.bin_info(column_x),
        **bin_y.bin_info(column_y),
        **{
            "frequency": [len(bin_y.filter(bin_x.filter(df, column_x), column_y))]
        }
    })


class QuantitativeBin(object):
    def __init__(self, lb, ub, is_last_bin):
        self.lb = lb
        self.ub = ub
        self.is_last_bin = is_last_bin

    def filter(self, df, column_name=""):
        if column_name:
            return df[(df[column_name]>=self.lb)&(df[column_name]<=self.ub)] if self.is_last_bin \
                else df[(df[column_name]>=self.lb)&(df[column_name]<self.ub)]
        else:
            return df[(df>=self.lb)&(df<=self.ub)] if self.is_last_bin \
                else df[(df>=self.lb)&(df<self.ub)]

    def bin_info(self, column_name=""):
        return {
            "{}_lb".format(column_name): [self.lb],
            "{}_ub".format(column_name): [self.ub]
        }


class CategoricalBin(object):
    def __init__(self, values):
        self.values = values

    def filter(self, df, column_name=""):
        return df[df[column_name].isin(self.values)] if column_name else df[df.isin(self.values)]

    def bin_info(self, column_name=""):
        return {
            column_name: self.values,
        } if len(self.values) == 1 else {column_name: ["OTHERS"]}


class BinCreationError(Exception):
    def __init__(self, message):
        self.message = message



import math
import copy
import logging

import numpy
import pandas
from pandas.api import types


logger = logging.getLogger(__name__)


def calc_column_stat(df):
    concat_list = []
    df_size = len(df)
    for column in df.columns:
        unique_count = len(get_unique_values(df, column))
        logger.info("...calculating {column} (dtype: {dtype})".format(
            column=column, dtype=df.dtypes[column]))
        concat_list.append(pandas.DataFrame({
            "column_name": [column],
            "dtype": df.dtypes[column],
            "min": [df[column].min()] if _orderable(df.dtypes[column]) else [pandas.NA],
            "max": [df[column].max()] if _orderable(df.dtypes[column]) else [pandas.NA],
            "mean": [df[column].mean()] if types.is_numeric_dtype(df.dtypes[column]) else [pandas.NA],
            "std": [df[column].std()] if types.is_numeric_dtype(df.dtypes[column]) else [pandas.NA],
            "ratio_na": [pandas.isnull(df[column]).sum() / float(len(df))],
            "ratio_zero": [len(df[df[column]==0]) / float(len(df))]\
                if types.is_numeric_dtype(df.dtypes[column]) else [pandas.NA],
            "unique_count": [unique_count],
            "is_unique": [unique_count == df_size]
        }, columns=["column_name", "dtype", "min", "max", "mean", "std", "ratio_na", "ratio_zero", "unique_count", "is_unique"]))
    return pandas.concat(concat_list, axis=0)


def get_unique_values(df, columns):
    df_tmp = df[columns].dropna()
    if isinstance(df_tmp, pandas.Series):
        return set([v for v in df_tmp.values])
    else:
        return set([tuple(v) for v in df_tmp.values])


def calculate_percentiles_for_df(df, column_list, ratio_list):
    value_dic = {"name": ["min"] +
                 ["{0:.2%}-percentile".format(ratio) for ratio in ratio_list]
                 + ["max"]}
    for column in column_list:
        value_dic[column] = calculate_percentiles(df[column].values, ratio_list)
    return pandas.DataFrame(value_dic, columns=["name"] + column_list).set_index("name")


def calculate_percentiles(array, ratio_list):
    n_record0 = array.shape[0]
    valid_array = array[numpy.isfinite(array)]
    n_record = valid_array.shape[0]
    if n_record0 - n_record > 0:
        logger.info("{0} records have missing values (out of {1} records)".format(n_record0 - n_record, n_record0))
    sorted_array = numpy.sort(valid_array)
    ind_list = [_get_ind(n_record, ratio) for ratio in ratio_list]
    return numpy.array([sorted_array[ind] for ind in ind_list])


def _get_ind(n_record, ratio):
    return 0 if ratio == 0 else int(math.ceil(n_record * ratio)) - 1


def _orderable(dtype):
    if types.is_numeric_dtype(dtype) or types.is_datetime64_any_dtype(dtype):
        return True
    else:
        return False



from scipy import stats
from sklearn import linear_model

from . import base


estimator = {
    "cl": linear_model.LogisticRegression(solver="liblinear"),
    "rg": linear_model.Lasso(),
    "mcl": linear_model.LogisticRegression(solver="liblinear")
}
grid_rg = dict(
    alpha=[0, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1.0, 10]
)
grid_cl = dict(
    penalty=["l1", "l2"],
    C=[1e-5, 1e-4, 1e-3, 1e-2, 1e-1, 1.0, 10.0]
)
distributions_rg = dict(
    alpha=stats.loguniform(1e-5, 10)
)
distributions_cl = dict(
    penalty=["l1", "l2"],
    C=stats.loguniform(1e-5, 10)
)
config = base.TunerConfig(estimator, distributions_rg, distributions_cl, grid_rg, grid_cl)



from datetime import datetime

UB_MISSING_RATIO=0.5
UB_UNIQUE_COUNT=1000


def alert_columns(stat_df):
    alert_list = []
    for index, record in stat_df.iterrows():
        alert_list.extend(_check_missing_ratio(record))
        if record["adt"] == "datetime": alert_list.extend(_check_datetime_range(record))
        if record["adt"] == "category": alert_list.extend(_check_unique_count(record))
        if record["adt"] == "numeric": alert_list.extend(_check_outliers(record))
    for alert in alert_list:
        display(alert)


def _check_unique_count(record):
    alert_list = [ _generate_alert_message(
                    "WARN", record["column_name"], "too many unique values {0}".format(record["unique_count"]))] \
                            if record["unique_count"] > UB_UNIQUE_COUNT else \
                            [_generate_alert_message("WARN", record["column_name"], "only single unique value")] \
                                    if record["unique_count"] == 1 else []
    return alert_list


def _check_datetime_range(record):
    alert_list = []
    if record["min"] < datetime.datetime(1900, 1, 1):
        alert_list.append(_generate_alert_message(
                "ERROR", record["column_name"], "too old timestamp value {0}".format(record["min"])))
    if record["max"] > datetime.datetime(2999, 12, 31):
        alert_list.append(_generate_alert_message(
                "ERROR", record["column_name"], "too far future timestamp value {0}".format(record["min"])))
    return alert_list 


def _check_missing_ratio(record):
    alert_list = [_generate_alert_message(
            "WARN", record["column_name"], "too many missing values {0}".format(record["ratio_na"]))] \
                    if record["ratio_na"] > UB_MISSING_RATIO else []
    return alert_list


def _check_outliers(record):
    return []


def _generate_alert_message(error_type, column, message):
    return "[{0}] column {1} : {2}".format(error_type, column, message)



import logging

import numpy
import pandas
from pandas.api import types


logger = logging.getLogger(__name__)


def read_csv(filepath_or_buffer, **kwargs):
    """
    pandas.read_csv with automatic data type inference for int / timestamp, and resolve string issue
    https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.errors.DtypeWarning.html
    """
    df = pandas.read_csv(filepath_or_buffer, **kwargs)
    df = convert_integer_columns(df)
    df = convert_timestamp_columns(df)
    df = convert_string_columns(df)
    return df


def convert_integer_columns(df):
    for c in df.columns:
        if numpy.issubdtype(df.dtypes[c], numpy.number):
            integer_series = pandas.Series(df[c], dtype="Int64")
            if numpy.equal(abs(integer_series - df[c]).max(), 0):
                df[c] = integer_series
    return df


def convert_timestamp_columns(df):
    timestamp_columns = get_timestamp_columns(df)
    for c in timestamp_columns:
        df[c] = pandas.to_datetime(df[c])
    return df


def convert_string_columns(df):
    for c in [c for c in df.columns if types.is_object_dtype(df.dtypes[c])]:
        df[c] = df[c].fillna("").astype(str).replace("", pandas.NA)
    return df


def get_timestamp_columns(df):
    date_columns = []
    for c in [c for c in df.columns if types.is_object_dtype(df.dtypes[c])]:
        series = df[~df[c].isnull()][c].head()
        try:
            pandas.to_datetime(series)
        except Exception:
            continue
        else:
            logger.debug("column {} is detected as timestamp: {}".format(c, series))
            date_columns.append(c)
    return date_columns


def _is_integer(series):
    if abs(pandas.Series(series, dtype="Int64") - series).max() == 0:
        return True



import pytest

import pandas

from conjurer import eda


@pytest.fixture
def input_df():
    input_df = pandas.DataFrame({
        "integer_1": [1, 2, -3, 0],
        "integer_2": [1, 2, 1, 2],
        "integer_with_null": [1, None, None, None],
        "float_1": [1.5, 2.5, -0.0, 1.3],
        "float_2": [0.1, 0.1, 0.1, 0.3],
        "float_with_null": [None, 0.5, None, -2.0],
        "string_1": ["a", "b", pandas.NA, "d"],
        "string_2": [pandas.NA, "a", "a", "a"],
        "string_with_null": ["aaa", pandas.NA, None, None],
        "all_null": [None, None, None, None]
    })
    for c in input_df.columns:
        dtype = "Int64" if c.startswith("int") else "float" if c.startswith("float") else "object"
        input_df[c] = input_df[c].astype(dtype)
    return input_df


def test_integer_1(input_df):
    _test(input_df, "integer_1", {1, 2, -3, 0})


def test_integer_2(input_df):
    _test(input_df, "integer_2", {1, 2})


def test_integer_with_null(input_df):
    _test(input_df, "integer_with_null", {1})


def test_float_1(input_df):
    _test(input_df, "float_1", {1.5, 2.5, -0.0, 1.3})


def test_float_2(input_df):
    _test(input_df, "float_2", {0.1, 0.3})


def test_float_with_null(input_df):
    _test(input_df, "float_with_null", {0.5, -2.0})


def test_string_1(input_df):
    _test(input_df, "string_1", {"a", "b", "d"})


def test_string_2(input_df):
    _test(input_df, "string_2", {"a"})


def test_string_with_null(input_df):
    _test(input_df, "string_with_null", {"aaa"})


def test_all_null(input_df):
    _test(input_df, "all_null", set([]))


def test_2_composite(input_df):
    _test(input_df, ["integer_1", "float_1"], {(1, 1.5), (2, 2.5), (-3, -0.0), (0, 1.3)})


def test_2_composite_with_null(input_df):
    _test(input_df, ["float_2", "string_with_null"], {(0.1, "aaa")})


def test_2_composite_all_null(input_df):
    _test(input_df, ["float_2", "all_null"], set([]))


def test_3_composite(input_df):
    _test(input_df, ["integer_1", "float_1", "string_with_null"], {(1, 1.5, "aaa")})


def _test(input_df, columns, expected_values):
    unique_values = eda.get_unique_values(input_df, columns)
    assert unique_values == expected_values, \
        "Unique values of {} should be {} (returned {})".format(columns, expected_values, unique_values)



import io
from datetime import datetime

import pandas
from pandas import testing

from conjurer import eda
from conjurer.logic.eda.load import pandas_csv
from . import csv_data


def test_integer():
    df = eda.read_csv(io.StringIO(csv_data.INTEGER_TEST_CSV))
    testing.assert_frame_equal(
        df,
        pandas.DataFrame({
            "int1": pandas.Series([1, 2, None, 3], dtype="Int64"),
            "int2": pandas.Series([None, 1, 2, 3], dtype="Int64")
        })
    )


def test_all_type():
    df = eda.read_csv(io.StringIO(csv_data.ALL_TYPE_TEST_CSV))
    testing.assert_frame_equal(
        df,
        pandas.DataFrame({
            "date1": pandas.Series([datetime(2020, 1, 1), datetime(2020, 12, 31),
                                    datetime(2000, 3, 1), datetime(2000, 3, 1)], dtype="datetime64[ns]"),
            "date2": pandas.Series([None, datetime(2020, 1, 1),
                                    datetime(2020, 3, 31), datetime(2020, 7, 30)], dtype="datetime64[ns]"),
            "timestamp1": pandas.Series([datetime(2020, 1, 1, 12, 12, 30), datetime(2020, 1, 1, 0, 0, 59),
                                         datetime(2020, 1, 1, 0, 0, 59), datetime(2020, 1, 1, 0, 0, 59)],
                                        dtype="datetime64[ns]"),
            "int1": pandas.Series([None, 1, 2, 3], dtype="Int64"),
            "float1": pandas.Series([1.1, 2.5, 3.8, -1.3], dtype="float64"),
            "str1": pandas.Series(["A-1113", "B-1515", pandas.NA, "C-1335"]),
            "null": pandas.Series([None, None, None, None], dtype="float64")
        })
    )


def test_get_timestamp_columns():
    date_columns = pandas_csv.get_timestamp_columns(pandas.read_csv(io.StringIO(csv_data.ALL_TYPE_TEST_CSV)))
    assert date_columns == ["date1", "date2", "timestamp1"]



import pandas
from numpy import random

from conjurer import eda


def test_integer_over_num_bins_array():
    series = pandas.Series(random.choice(range(100), 10000), dtype="Int64")
    eda.plot_histogram(series, num_bins=11).display()


def test_integer_under_num_bins_array():
    series = pandas.Series(random.choice(range(10), 10000), dtype="Int64")
    eda.plot_histogram(series, num_bins=11).display()



import logging

import altair as alt
from pandas.api import types

from conjurer.logic.eda.vis import (
    binned_bar,
    binning
)


logger = logging.getLogger(__name__)


def plot_histogram(values, num_bins=50, normalize=False, minv=None, maxv=None):
    try:
        bin_df = binning.create_frequency_table(values, num_bins, minv, maxv)
    except Exception as e:
        raise e
    column_y = "ratio" if normalize else "frequency"
    if len(bin_df.columns) > 3:
        return binned_bar.plot_bar_with_binned(bin_df, bin_df.columns[0], bin_df.columns[1], column_y)
    else:
        args = {} if types.is_integer_dtype(values.dtype) else dict(sort="-y")
        return alt.Chart(bin_df).mark_bar().encode(
            x=alt.X("{}:N".format(bin_df.columns[0]), **args),
            y=column_y
        ).properties(height=200, width=800).interactive()


def plot_histogram_for_stats(df, stat_df, num_bins=50, normalize=False):
    for ind, column in enumerate(df.columns):
        minv, maxv = \
            stat_df[stat_df["column_name"] == column][
                ["min", "max"]].values.tolist()[0]
        try:
            plot_histogram(df[column], num_bins, normalize, minv, maxv).display()
        except binning.BinCreationError as e:
            logger.info("Histogram for {} was skipped: {}".format(column, e.message))
            pass



import logging
from functools import reduce
from operator import mul

import numpy
from sklearn import model_selection

from . import model


logger = logging.getLogger(__name__)


class RandomizedSearchCV(model_selection.RandomizedSearchCV):
    """
    sklearn.model_selection.RandomizedSearchCV with pandas DataFrame interface
    """
    def __init__(self, *args, **kwargs):
        """
        The same manner as [sklearn.model_selection.RandomizedSearchCV](https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html)
        """
        kwargs["return_train_score"] = True
        kwargs["verbose"] = 10
        super(RandomizedSearchCV, self).__init__(*args, **kwargs)

    def fit_sv_pandas(self, df_training, target_column, feature_columns,
                      df_validation=None, ratio_training=None, **kwargs):
        """
        `fit` for pandas DataFrame to perform single validation
        Args:
            df_training (pandas.DataFrame): training data set
            target_column (str): column name of prediction target
            feature_columns (list of str): column names of features
            df_validation (pandas.DataFrame): if specified, used as validation data set
            ratio_training (float): if specified, `df_training` is split for training / validation
            **kwargs: Other keyword arguments for original `fit`

        Returns:
            conjurer.ml.Model
        """
        x, y, num_training, num_validation = _split_for_sv(
            df_training, target_column, feature_columns, df_validation, ratio_training)
        self.cv = model_selection.PredefinedSplit(
            numpy.array([-1] * num_training + [0] * num_validation))
        logger.warning("start learning with {} hyper parameters".format(self.n_iter))
        self.fit(x, y, **kwargs)
        return model.Model(self, feature_columns=feature_columns, target_column=target_column)

    def fit_cv_pandas(self, df, target_column, feature_columns, n_fold, **kwargs):
        """
        `fit` for pandas DataFrame to perform cross validation
        Args:
            df (pandas.DataFrame): training data set
            target_column (str): column name of prediction target
            feature_columns (list of str): column names of features
            n_fold (int): The number of fold for CV
            **kwargs: Other keyword arguments for original `fit`

        Returns:
            conjurer.ml.Model
        """
        df = df.sample(len(df))  # shuffle
        x = df[feature_columns].values
        y = df[target_column].values
        self.cv = n_fold
        logger.warning("start learning with {} hyper parameters".format(self.n_iter))
        self.fit(x, y, **kwargs)
        return model.Model(self, feature_columns=feature_columns, target_column=target_column)


class GridSearchCV(model_selection.GridSearchCV):
    def __init__(self, *args, **kwargs):
        kwargs["return_train_score"] = True
        kwargs["verbose"] = 10
        super(GridSearchCV, self).__init__(*args, **kwargs)

    def fit_sv_pandas(self, df_training, target_column, feature_columns,
                      df_validation=None, ratio_training=None, **kwargs):
        """
        `fit` for pandas DataFrame to perform single validation
        Args:
            df_training (pandas.DataFrame): training data set
            target_column (str): column name of prediction target
            feature_columns (list of str): column names of features
            df_validation (pandas.DataFrame): if specified, used as validation data set
            ratio_training (float): if specified, `df_training` is split for training / validation
            **kwargs: Other keyword arguments for original `fit`

        Returns:
            conjurer.ml.Model
        """
        x, y, num_training, num_validation = _split_for_sv(
            df_training, target_column, feature_columns, df_validation, ratio_training)
        self.cv = model_selection.PredefinedSplit(
            numpy.array([-1] * num_training + [0] * num_validation))
        logger.warning("start learning with {} parameters".format(_get_num_parameters(self.param_grid)))
        self.fit(x, y, **kwargs)
        return model.Model(self, feature_columns=feature_columns, target_column=target_column)

    def fit_cv_pandas(self, df, target_column, feature_columns, n_fold, **kwargs):
        """
        `fit` for pandas DataFrame to perform cross validation
        Args:
            df (pandas.DataFrame): training data set
            target_column (str): column name of prediction target
            feature_columns (list of str): column names of features
            n_fold (int): The number of fold for CV
            **kwargs: Other keyword arguments for original `fit`

        Returns:
            conjurer.ml.Model
        """
        df = df.sample(len(df))  # shuffle
        x = df[feature_columns].values
        y = df[target_column].values
        self.cv = n_fold
        logger.warning("start learning with {} parameters".format(_get_num_parameters(self.param_grid)))
        self.fit(x, y, **kwargs)
        return model.Model(self, feature_columns=feature_columns, target_column=target_column)


def _split_for_sv(df_training, target_column, feature_columns, df_validation, ratio_training):
    if df_validation is not None:
        x = numpy.concatenate(
            (df_training[feature_columns].values, df_validation[feature_columns].values),
            axis=0
        )
        y = numpy.concatenate(
            (df_training[target_column].values, df_validation[target_column].values),
            axis=0
        )
        num_training = len(df_training)
        num_validation = len(df_validation)
    else:
        shuffled_df = df_training.sample(len(df_training))
        x = shuffled_df[feature_columns].values
        y = shuffled_df[target_column].values
        num_training = int(ratio_training * len(df_training))
        num_validation = len(df_training) - num_training
    return x, y, num_training, num_validation


def _get_num_parameters(param_grid):
    product = lambda list_values: reduce(mul, list_values, 1)
    return len(param_grid) if isinstance(param_grid, list) \
        else product([len(param_grid[elem]) for elem in param_grid.keys()])



import altair as alt


def plot_bar_with_binned(df, column_lb, column_ub, column_y, xname=None):
    xname = xname or column_lb.replace("_lb", "")
    return alt.Chart(df).mark_bar().encode(
        alt.X(column_lb, bin="binned", axis=alt.Axis(title=xname)),
        x2=column_ub,
        y=alt.Y("{}:Q".format(column_y)),
        tooltip=[column_lb, column_ub, column_y]
    ).properties(height=200, width=800).interactive()





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="conjurer",
    version="0.0.12",
    author="@not-so-fat",
    author_email="conjurer.not.so.fat@gmail.com",
    description="Python library to help you to perform magic on your data analytics project",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/not-so-fat/conjurer",
    install_requires=[
        "pandas>=1.1", "numpy", "scipy", "sklearn", "altair", "IPython",
        "lightgbm", "xgboost", "gbm_autosplit>=0.0.1", "networkx"
    ],
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import numpy
from sklearn import utils

from .logic.ml.tuner import (
    cv_analyzer,
    ml_tuner
)
from .logic.ml.sklearn_cv_pandas import pandas_cv


RandomizedSearchCV = pandas_cv.RandomizedSearchCV
GridSearchCV = pandas_cv.GridSearchCV
CVAnalyzer = cv_analyzer.CVAnalyzer


def get_default_cv(ml_type, problem_type, scoring=None, search_type="random"):
    """
    Get pandas-I/F [Randomized|Grid]SearchCV object with default search parameters and preprocessing
    Args:
        ml_type (str): One of "lightgbm", "xgboost", "random_forest", "linear_model"
        problem_type (str): One of "cl", "rg"
        scoring (str or scorer): `scoring` for RandomizedSearchCV / GridSearchCV
        search_type (str): "random" to use "RandomizedSearchCV", otherwise used "GridSearchCV"
        param_dict (str): parameter search space
    Returns:
        RandomizedSearchCV or GridSearchCV
    """
    return ml_tuner.get_cv(ml_type, problem_type, scoring, search_type)


def estimate_std(actual, predicted, metric, n_bootstrap=10, n_samples=None):
    """
    Estimate standard deviation for the metric in specified dataset by bootstrapping
    Args:
        actual(numpy.array):
        predicted(numpy.array):
        metric:
        n_bootstrap:
        n_samples:

    Returns:

    """
    return numpy.std(numpy.array([
        metric(*utils.resample(actual, predicted, n_samples=n_samples))
        for i_bootstramp in range(n_bootstrap)
    ]))
